%% Project: MATLAB supported Synchro-measurement Application Development Framework (SADF)
% Author: Matija Naglic
% Email: m.naglic@tudelft.nl
% Version: 2.0
% Date: 08/01/2018
% Reference paper: M. Naglic, M. Popov, M. Van Der Meijden, V. Terzija, "Synchro-measurement Application Development Framework: an IEEE Standard C37.118.2-2011 Supported MATLAB Library", IEEE Transactions on Instrumentation & Measurements
% Reference paper DOI: https://doi.org/10.1109/TIM.2018.2807000

% // Copyright (C) 2018   Matija Naglic     <m.naglic@tudelft.nl>
% //
% // This program is free software; you can redistribute it and/or modify
% // it under the terms of the GNU General Public License as published by
% // the Free Software Foundation; either version 3 of the License, or
% // (at your option) any later version.
% //
% // This program is distributed in the hope that it will be useful,
% // but WITHOUT ANY WARRANTY; without even the implied warranty of
% // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% // GNU General Public License for more details.
% //
% // You should have received a copy of the GNU General Public License
% // along with this program; if not, see <http://www.gnu.org/licenses/>.


%% iefls, Online PMU monitoring
function IUFLS
global DATA CFG_2_3 iefls

if ~isfield(iefls,'action')   %iefls VAR initialisation
    iefls.action = false(1,3);
    iefls.action_location =  zeros(1,3);
    iefls.counter =  zeros(1,3);
    iefls.reset = false;
    iefls.SKT_IP='131.180.164.23';
    iefls.SKT_Port = 4700;
    iefls.Skt_Dev = tcpclient(iefls.SKT_IP, iefls.SKT_Port, 'Timeout', 2);
    iefls.CB_trip=zeros(1,3);
    
    figure('units', 'normalized', 'outerposition', [0 0 1 1], 'name','Synchro-measurement Application Development Framework');
    iefls.window = 250; %display_samples
    iefls.phasor = 1;
    iefls.processed = 0;
end

if DATA.index_max >  iefls.processed %Check for new data
    
    iefls.processed = DATA.index_max;
    
    if  any(DATA.FreqDev(DATA.index_max,1,:) <= -0.8)  &&   iefls.action(1) == false
        disp([datestr(now, 'yyyy-mm-dd HH:MM:SS.FFF') '   UFLS action 1']);
        iefls.reset = false;
        iefls.action(1) = 1;
        iefls.counter(1) = iefls.counter(1) +1;
        iefls.action_location(1) = DATA.index_max;
        iefls.CB_trip(1, 1) = 1;
        write(iefls.Skt_Dev, typecast(swapbytes(uint32(iefls.CB_trip)), 'uint8'));
        
    elseif  any(DATA.FreqDev(DATA.index_max,1,:) <= -1.2)  &&   iefls.action(2) == false  &&   iefls.action(1) == true
        disp([datestr(now, 'yyyy-mm-dd HH:MM:SS.FFF') '   UFLS action 2']);
        iefls.action(2) = 1;
        iefls.counter(2) = iefls.counter(2) +1;
        iefls.action_location(2) = DATA.index_max;
        iefls.CB_trip(1, 2) = 1;
        write(iefls.Skt_Dev, typecast(swapbytes(uint32(iefls.CB_trip)), 'uint8'));
        
    elseif  any(DATA.FreqDev(DATA.index_max,1,:) <= -2)  &&   iefls.action(3) == false &&   iefls.action(2) == true
        disp([datestr(now, 'yyyy-mm-dd HH:MM:SS.FFF') '   UFLS action 3']);
        iefls.action(3) = 1;
        iefls.counter(3) = iefls.counter(3) +1;
        iefls.action_location(3) = DATA.index_max;
        iefls.CB_trip(1, 3) = 1;
        write(iefls.Skt_Dev, typecast(swapbytes(uint32(iefls.CB_trip)), 'uint8'));
        
        
    elseif any(DATA.FreqDev(DATA.index_max,1,:) >= 0.5)  &&  ~iefls.reset   % any(iefls.action == true) && any(DATA.index_max > (iefls.action_location + 60 * CFG_2_3.Data_Rate))
        disp([datestr(now, 'yyyy-mm-dd HH:MM:SS.FFF') '   UFLS reset']);
        iefls.CB_trip=zeros(1,3);
        iefls.action = false(1,3);
        iefls.reset = true;
        write(iefls.Skt_Dev, typecast(swapbytes(uint32(iefls.CB_trip)), 'uint8'));
    end
    
    
    if 1  %enable/disable figure update
        clf;
        hold on;
        title_plot=[];
        
        for i = 1:CFG_2_3.Num_PMU(CFG_2_3.counter, 1)
            if DATA.index_max <= iefls.window
                iefls.dataset = [DATA.TimeStamp(1:DATA.index_max,1) DATA.Freq(1:DATA.index_max,iefls.phasor,i)];
            else
                iefls.dataset = [DATA.TimeStamp(DATA.index_max-iefls.window:DATA.index_max,1) DATA.Freq(DATA.index_max-iefls.window:DATA.index_max,iefls.phasor,i)];
            end
            stairs(iefls.dataset(~isnan(iefls.dataset(:,1)),1), iefls.dataset(~isnan(iefls.dataset(:,1)),2));
            title_plot = [title_plot strtrim(char(CFG_2_3.Station_Name(:, i))') ' - ' strtrim(char(CFG_2_3.PMUsPhasorName(:, iefls.phasor, i))') ', '];
        end
        
        ax = gca;
        ax.XTick = iefls.dataset(~isnan(iefls.dataset(:,1)),1);
        xlabel('timestamp (UTC)'); ylabel('system frequency (Hz)');
        title(['Online Monitoring: ' title_plot ])
        grid on
        datetick('x', 'dd-mm-yyyy HH:MM:SS.FFF', 'keeplimits', 'keepticks')
        set(gca, 'XMinorTick', 'on', 'YMinorTick', 'on', 'XTickLabelRotation', 45)
        set(findall(gca, 'Type', 'Line'),'LineWidth',1);
        drawnow
        % disp([datestr(now, 'yyyy-mm-dd HH:MM:SS.FFF') '  PLOT ']);
    end
end
end
